/* gcc http.c -o http */

/* stdlib */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* socket */
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>

/* signal */
#include <signal.h>

void handle(int connection)
{
    char buffer[1024 * 8]; /* - 8kb buffer */

    /* receive */
    size_t read = recv(connection, buffer, sizeof(buffer) / sizeof(buffer[0]), MSG_NOSIGNAL);

    /* isolate requested path */
    char *path = buffer;
    int i = 0;
    while (i++ < read && *path++ != ' ');   /* find start of path */
    while (i++ < read) {                    /* find end of path */
        if (path[i] == ' ') {
            path[i] = '\0';
            break;
        }
    }
    if (path[i - 1] == '/') { /* redirect from root directory */
        const char headerBegin[] = "HTTP/1.1 301 Moved Permanently\nLocation: ";
        const char headerEnd[] = "index.html\n";
        send(connection, headerBegin, strlen(headerBegin), MSG_NOSIGNAL);
        send(connection, path, strlen(path), MSG_NOSIGNAL);
        send(connection, headerEnd, strlen(headerEnd), MSG_NOSIGNAL);
        return;
    }
    /* make path relative */
     path--;
    *path = '.';

    /* respond */
    FILE *file = fopen(path, "r");
    if (file) {
        const char header[] = "HTTP/1.1 200 OK\n\n";
        if (send(connection, header, strlen(header), MSG_NOSIGNAL) > 0) {
            while (   (read = fread(buffer, 1, sizeof(buffer) / sizeof(buffer[0]), file)) > 0
                    && send(connection, buffer, read, MSG_NOSIGNAL) == read);
        }
        fclose(file);
    } else {
        const char header[] = "HTTP/1.1 404 Not Found\n";
        send(connection, header, strlen(header), MSG_NOSIGNAL);
    }
}

static int volatile fd = 0;

int create_socket(int port)
{
    int fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        printf("socket() failed\n");
        return 0;
    }

    struct sockaddr_in hints = { 0 };
    hints.sin_family = AF_INET;
    hints.sin_port = htons(port);                 /* port number */
    hints.sin_addr.s_addr = htonl(INADDR_ANY);    /* internet address */

    if (bind(fd, (struct sockaddr *)&hints, sizeof(struct sockaddr_in)) < 0) {
        printf("bind() failed\n");
        close(fd);
        return 0;
    }
    if (listen(fd, 1) < 0) {
        printf("listen() failed\n");
        close(fd);
        return 0;
    }

    return fd;
}

static int volatile running = 1;

void sigint_handler(int signal) {
    running = 0;
    shutdown(fd, SHUT_RD);
}

int main()
{
    signal(SIGINT, sigint_handler);

    int err = 1;

    if ((fd = create_socket(8083))) {
        while (running) {
            struct sockaddr_in client;
            socklen_t sockaddr_len = sizeof(struct sockaddr_in);
            /* accept socket */
            int connection = accept(fd, (struct sockaddr *)&client, &sockaddr_len);
            if (connection > 0) {
                handle(connection);
                close(connection);
            }
        }

        err = 0;

        close(fd);
    }
    return err;
}
